/* Top-level object parsing.
   Copyright (C) 2001 Michal Moskal (malekith@pld.org.pl)

This file is part of GCC.

GCC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GCC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */


#include "ksi.h"

tree bool_type_node;

const char *storage_name(storage_t s)
{
	switch (s) {
	case storage_static: return "static";
	case storage_public: return "public";
	case storage_extern: return "extern";
	default: return "BAD_STORAGE";
	}
}

storage_t lookup_storage(const char *s)
{
	storage_t r = storage_error;
	
	switch (s[0]) {
	case 's':
		r = storage_static;
		break;
	case 'p':
		r = storage_public;
		break;
	case 'e':
		r = storage_extern;
		break;
	}

	if (r == storage_error || strcmp(storage_name(r), s) != 0)
		fatal_error("bad storage specifier '%s'", s);
	
	return r;
}

const char *type_name(type_t t)
{
	switch (t) {
	case int_type: return "int";
	case uint_type: return "uint";
	case float_type: return "float";
	case ptr_type: return "ptr";
	case void_type: return "void";
	case bool_type: return "bool";
	default: return "BAD_TYPE";
	}
}

type_t lookup_type(const char *type)
{
	type_t t = bad_type;
	
	switch (*type) {
	case 'i': t = int_type; break;
	case 'u': t = uint_type; break;
	case 'f': t = float_type; break;
	case 'p': t = ptr_type; break;
	case 'v': t = void_type; break;
	case 'b': t = bool_type; break;
	}
	
	if (t == bad_type || strcmp(type_name(t), type) != 0)
		fatal_error("bad type '%s'", type);
	
	return t;
}

/* type(int|uint|float|ptr|void [,SIZE]) */
tree parse_type(void)
{
	struct term_state ctx;
	struct arg a;
	int size = 0;
	type_t t;

	init_term(&ctx, kw_type);
	get_term_arg(&ctx, arg_id, &a);
	t = lookup_type(a.u.val_id);
	free_arg(&a);

	if (optional_term_arg_present(&ctx)) {
		get_term_arg(&ctx, arg_int|arg_uint, &a);
		size = a.u.val_int.low;
		free_arg(&a);
	}
	
	fini_term(&ctx);

	if (((t == int_type || t == uint_type) &&
	       (size != 8 && size != 16 && size != 32 && 
	        size != 64 && size != 128)) ||
	    ((t == float_type) &&
	     (size != FLOAT_TYPE_SIZE && 
	      size != DOUBLE_TYPE_SIZE && 
	      size != LONG_DOUBLE_TYPE_SIZE)) ||	/* hmm.. i'm not sure */
	    ((t == ptr_type || t == void_type) && size != 0))
	     	fatal_error("%s type size cannot be %d", type_name(t), size);
	
	switch (t) {
	case int_type:
		return type_for_size(size, 0);
	case uint_type:
		return type_for_size(size, 1);
	case float_type:
		switch (size) {
		case FLOAT_TYPE_SIZE:
			return float_type_node;
		case DOUBLE_TYPE_SIZE:
			return double_type_node;
		default:
			/* non-integer constant */
			if (size == LONG_DOUBLE_TYPE_SIZE)
				return long_double_type_node;
			ice();
		}
	case void_type:
		return void_type_node;
	case ptr_type:
		return ptr_type_node;
	case bool_type:
		return bool_type_node;
	default:
		ice();
	}

	return NULL;
}

/* var(type(uint, 32), x) */
void parse_var(tree *type, tree *name)
{
	struct term_state ctx;
	struct arg a;

	init_term(&ctx, kw_var);

	expect_keyword(&ctx, kw_type);
	*type = parse_type();

	get_term_arg(&ctx, arg_id, &a);
	*name = get_identifier(a.u.val_id);
	free_arg(&a);

	fini_term(&ctx);
}

static void parse_func_args_r(struct list_state *ctx, tree *param_list, tree *param_type_list)
{
	keyword_t kw = get_keyword_from_list(ctx);
	tree type, name, arg;
	
	switch (kw) {
	case kw_nil:
		arg0_term(kw);
		if ((kw = get_keyword_from_list(ctx)) != kw__error)
			fatal_error("found (%s ...) aftrer (nil) in parm list", 
				    keyword_name(kw));
		*param_type_list = tree_cons(NULL_TREE, void_type_node, NULL_TREE);
		break;
	case kw_var:
		parse_var(&type, &name);
		arg = build_decl(PARM_DECL, name, type);
		DECL_ARG_TYPE(arg) = type;
		/* we need to link this in reverse order */
		parse_func_args_r(ctx, param_list, param_type_list);
		*param_list = chainon(arg, *param_list);
		*param_type_list = tree_cons(NULL_TREE, type, *param_type_list);
		break;
	case kw_elipsis:
		arg0_term(kw);
		if ((kw = get_keyword_from_list(ctx)) != kw__error)
			fatal_error("found (%s ...) aftrer (elipsis) in parm list", 
				    keyword_name(kw));
		*param_type_list = NULL_TREE;
		break;
	case kw__error:
		*param_type_list = tree_cons(NULL_TREE, void_type_node, NULL_TREE);
		break;
	default:
		fatal_error("found (%s ...) where (var ...) or (nil) was expected",
			    keyword_name(kw));
	
	}
}

static void parse_func_args(tree *param_list, tree *param_type_list)
{
	struct list_state ctx;

	init_list(&ctx, kw_parm);
	parse_func_args_r(&ctx, param_list, param_type_list);
}

static void compile_body(tree fndecl, keyword_t kw)
{
	tree param_decl, next_param;
	tree body, dummy;

	/* Set line number information; everything is line 1 for us.  */
	DECL_SOURCE_FILE(fndecl) = input_filename;
	DECL_SOURCE_LINE(fndecl) = 1;

	/* Announce we are compiling this function.  */
	announce_function(fndecl);

	/* Set up to compile the function and enter it.  */
	current_function_decl = fndecl;
	DECL_INITIAL(fndecl) = error_mark_node;

	pushlevel(0);
	make_decl_rtl(fndecl, 0);

	/* Push all the PARM_DECL nodes onto the current scope (i.e. the scope of the
	   subprogram body) so that they can be recognized as local variables in the
	   subprogram.   */

	for (param_decl = nreverse(DECL_ARGUMENTS(fndecl));
	     param_decl; param_decl = next_param) {
		next_param = TREE_CHAIN(param_decl);
		TREE_CHAIN(param_decl) = NULL;
		pushdecl(param_decl);
	}

	/* Store back the PARM_DECL nodes. They appear in the right order. */
	DECL_ARGUMENTS(fndecl) = getdecls();

	init_function_start(fndecl, input_filename, 1);
	expand_function_start(fndecl, 0);
	expand_start_bindings(0);

	/* The only thing in this function is a single return statement, which
	   we now generate.  Note that EXP will have been allocated in
	   permanent_obstack.  To do this properly, we need to enter the function 
	   ontext before parsing the expression, but then error recovery is more
	   complex, so we don't do that here.  */

/* expand_return (build (MODIFY_EXPR, void_type_node, */
/* DECL_RESULT (fndecl), exp)); */

	if (kw != kw_bind)
		fatal_error("function body top-expressoion isn't "
			    "(bind ...), but (%s ...)", 
			    keyword_name(kw));
			    
	body = compile(kw, &dummy);

	fini_labels();
	
	if (flag_yydebug)
		debug_tree(body);

	expand_expr_stmt(body);

	/* Now get back out of the function and compile it.  */
	expand_end_bindings(NULL_TREE, 1, 0);

	poplevel(1, 0, 1);

	BLOCK_SUPERCONTEXT(DECL_INITIAL(fndecl)) = fndecl;

	/* Generate rtl for function exit.  */
	expand_function_end(input_filename, lineno, 0);

	/* Run the optimizers and output assembler code for this function. */
	rest_of_compilation(fndecl);

	/* Garbage collect */
	ggc_collect();

	current_function_decl = NULL_TREE;
}

/* func(static|public|extern, RET_TYPE, NAME, ARGS, BODY) */
void parse_func(void)
{
	struct term_state ctx;
	struct arg a;
	storage_t storage;
	keyword_t kw;
	tree ret_type;
	tree name;
	tree fndecl, fntype;
	tree param_list = NULL_TREE;
	tree param_type_list = NULL_TREE;
	
	init_term(&ctx, kw_func);
	
	get_term_arg(&ctx, arg_id, &a);
	storage = lookup_storage(a.u.val_id);
	free_arg(&a);
	
	expect_keyword(&ctx, kw_type);
	ret_type = parse_type();
	
	get_term_arg(&ctx, arg_id, &a);
	name = get_identifier(a.u.val_id);
	free_arg(&a);
	
	parse_func_args(&param_list, &param_type_list);

	/* The function type depends on the return type and type of args.  */
	fntype = build_function_type(ret_type, param_type_list);

	/* first keyword of the body */
	kw = get_keyword(&ctx);
	
	/* Now make the function decl.  */
	fndecl = build_decl(FUNCTION_DECL, name, fntype);
	DECL_EXTERNAL(fndecl) = storage == storage_extern;
	TREE_PUBLIC(fndecl) = storage != storage_static;
	TREE_STATIC(fndecl) = kw == kw_nil;
	DECL_ARGUMENTS(fndecl) = param_list;
	DECL_RESULT(fndecl) = build_decl(RESULT_DECL, NULL_TREE, ret_type);
	DECL_CONTEXT(DECL_RESULT(fndecl)) = fndecl;

	rest_of_decl_compilation(fndecl, NULL, 1, 0);

	pushdecl(fndecl);
	
	if (kw == kw_nil) {
		arg0_term(kw);
	} else if (storage == storage_extern)
		fatal_error("extern function supplaied with body");
	else
		compile_body(fndecl, kw);
		
	fini_term(&ctx);
}

void parse_global()
{
	sorry("global");
	exit(1);
}

void parse_unit()
{
	struct list_state ctx;
	keyword_t kw;

	init_list(&ctx, kw_unit);
	
	while ((kw = get_keyword_from_list(&ctx)) != kw__error) {
		switch (kw) {
		case kw_func:
			parse_func();
			break;
		case kw_global:
			parse_global();
			break;
		default:
			fatal_error("(%s ...) cannot be part of unit", keyword_name(kw));
			break;
		}
	}
}
